/*
 * CurvesView.m
 * 2/25/99 Mike Trent
 *
 * A nifty bezier curve module. I feel I ought to explain one of the data
 * structures in detail. Each curve is represented by 4 points. These points
 * are stored in an array of NSRect structs:
 *
 *	_pts[lineNumber][points 1 - 4]
 *
 * The actual point is the NSRect's origin field. The NSRect's size field
 * stores the offset by which to move the point. For example, the size's width
 * is added to the point's x component to move the point. One should resist the
 * urge to treat these NSRects as NSRects; as the size can have negative
 * values, some Foundation functions will choke one them.
 */

#import "CurvesView.h"

float randBetween(float a, float b);

static float fland(float a, float b)
{
    return floor(randBetween(a,b));
}

@implementation CurvesView

- (id)initWithFrame:(NSRect)frameRect
{
    self = [super initWithFrame:frameRect];

    if (self) {
        int j;
        
        [self allocateGState];
        srandom(time(0));

        // we build a color array for performance. This array is never freed.
        // It might be better to build and free this array every time screen
        // saver mode stops and starts.
        _color[0] = [[NSColor blackColor] retain];
        for (j = 0; j < kLength; j++) {
            _color[j+1] = [[NSColor colorWithCalibratedHue:(1.0 / kLength * j)
                                                saturation:1.0
                                                brightness:1.0
                                                     alpha:1.0] retain];
        }
        _bp = [[NSBezierPath bezierPath] retain];
    }

    return self;
}

- (void)setFrameSize:(NSSize)newSize
{
    NSRect r;
    int i,j;

    [super setFrameSize:newSize];

    // initialize our line arrays. The first line will have a value, but
    // remaining lines will be zeroed out.
    r = [self bounds];
    for (j = 0; j < kLength; j++) {
        for (i = 0; i < kPoints; i++) {
            if (j ==0) {
                _pts[j][i].origin = NSMakePoint(fland(0.0, r.size.width),
                                                fland(0.0, r.size.width));
                _pts[j][i].size = NSMakeSize([self randOffset],
                                          [self randOffset]);
            } else {
                _pts[j][i].origin = NSMakePoint(0.0, 0.0);
                _pts[j][i].size = NSMakeSize(0.0, 0.0);
            }
        }
    }
    _idx = 1;
}

- (void)drawRect:(NSRect)rects
{
    [[NSColor blackColor] set];
    NSRectFill(rects);
}

- (void)oneStep
{
    int i, last;

    // if the line at _idx exists, erase it. The line might not exist if
    // it were initialized to nil in setFrameSize. 
    if (_pts[_idx][0].size.width != 0.0){
        [_color[0] set];
        [_bp moveToPoint:_pts[_idx][0].origin];
        [_bp curveToPoint:_pts[_idx][1].origin
            controlPoint1:_pts[_idx][2].origin
            controlPoint2:_pts[_idx][3].origin];
        [_bp stroke];
        [_bp removeAllPoints];
    }

    // re-initialize _idx to the value of the previous line, and then
    // move _idx's points..
    if (_idx == 0) last = kLength - 1;
    else last = _idx - 1;
    for (i = 0; i < kPoints; i++) {
        _pts[_idx][i] = _pts[last][i];
    }
    [self movePoints];

    // stroke our new line.
    [_bp moveToPoint:_pts[_idx][0].origin];
    [_bp curveToPoint:_pts[_idx][1].origin
        controlPoint1:_pts[_idx][2].origin
        controlPoint2:_pts[_idx][3].origin];
    [_color[_idx+1] set];
    [_bp stroke];
    [_bp removeAllPoints];

    _idx++;
    if (_idx >= kLength) _idx = 0;
}

- (void)movePoints
{
    NSRect r = [self bounds];
    int i;

    // Move every point by a random offset. If the point hits the side of the
    // view, change the point's direction and random offset.
    for (i = 0; i < kPoints; i++) {
        NSRect p = _pts[_idx][i];
        float xs = (p.size.width >= 0) ? 1 : -1;
        float ys = (p.size.height >= 0) ? 1 : -1;

        if (p.origin.x + p.size.width > r.size.width) {
            p.origin.x = r.size.width;
            xs = -xs;
            p.size.width = [self randOffset] * xs;
            p.size.height = [self randOffset] * ys;
        } else if (p.origin.x + p.size.width < 0.0) {
            p.origin.x = 0.0;
            xs = -xs;
            p.size.width = [self randOffset] * xs;
            p.size.height = [self randOffset] * ys;
        } else {
            p.origin.x += p.size.width;
        }

        if (p.origin.y + p.size.height > r.size.height) {
            p.origin.y = r.size.height;
            ys = -ys;
            p.size.width = [self randOffset] * xs;
            p.size.height = [self randOffset] * ys;
        } else if (p.origin.y + p.size.height < 0.0) {
            p.origin.y = 0.0;
            ys = -ys;
            p.size.width = [self randOffset] * xs;
            p.size.height = [self randOffset] * ys;
        } else {
            p.origin.y += p.size.height;
        }

        _pts[_idx][i] = p;
    }

}

- (NSView *)inspector:sender
{
    if (!sharedInspectorPanel) {
        [NSBundle loadNibNamed:@"Curves.nib" owner:self];
    }

    return sharedInspectorPanel;
}

- (float)randOffset
{
    return fland(1.0, 20.0);
}

@end

@implementation StaticCurvesView

- (void)drawRect:(NSRect)r
{
    int i,j;

    [super drawRect:r];

    for (j = 0; j < kLength; j++) {
        for (i = 0; i < kPoints; i++) {
            if (j ==0) {
                _pts[j][i].origin = NSMakePoint(fland(0.0, r.size.width),
                                                fland(0.0, r.size.width));
                _pts[j][i].size = NSMakeSize([self randOffset],
                                             [self randOffset]);
            } else {
                _pts[j][i].origin = NSMakePoint(0.0, 0.0);
                _pts[j][i].size = NSMakeSize(0.0, 0.0);
            }
        }
    }
    _idx = 1;
    
    for (i = 0; i < kLength-1; i++) {
        [self oneStep];
    }
}

- (float)randOffset
{
    return fland(1.0, 5.0);
}

@end
