/*
 * FreeSpace.m
 * 7/?/98 Mike Trent
 *
 * FreeSpace screen saver module.
 *
 * 8/10/98  MDT added setFrameSize to better deal with BackSpace. Moved sizing code out of
 *              enteredScreenSaverMode and into seedImage, a new method. Rescan if controls
 *		change.
 */

#import "FreeSpaceView.h"
#ifdef GNUSTEP
#import <math.h>
#endif

static NSString *speedStrings[] = {
    @"Drifting",
    @"Impulse",
    @"Warp Factor 1",
    @"Warp Factor 2",
    @"10.0 Bytemarks",
};

static NSString *starsStrings[] = {
    @"Few",
    @"Some",
    @"Many",
    @"Lots",
    @"Gobs",
};

static NSSize starsSizes[] = {
    {1, 1},
    {1, 2},
    {2, 2},
    {2, 3},
    {3, 3},
    {3, 4},
    {4, 4},
};


float randBetween(float a, float b);

static NSPoint drawptForStarInRect(Star *s, NSRect r)
{
    NSPoint p;

    p.x = floor(r.size.width / 2 + (s->distance * s->costheta));
    p.y = floor(r.size.height / 2 + (s->distance * s->sintheta));

    return p;
}

@implementation FreeSpaceView

//
// Initialization
//

- (BOOL)useBufferedWindow
{
    return NO;
}

- (id)initWithFrame:(NSRect)frameRect
{
    return [self initWithFrame:frameRect loadInspectorPanel:YES];
}

- (id)initWithFrame:(NSRect)frameRect loadInspectorPanel:(BOOL)loadPanel
{
    if (![super initWithFrame:frameRect]) return nil;

    [self allocateGState];
    srandom(time(0));

    [self loadDefaults];

    if (loadPanel) {
        (void)[self inspector:self];

        // set defaults
        [speedSlider setIntValue:speed];
        [speedText setStringValue:speedStrings[speed-1]];
        [starsSlider setIntValue:stars];
        [starsText setStringValue:starsStrings[stars-1]];
        [colorBtn setState:color];
   }

    colors[0] = [[NSColor whiteColor] retain];
    colors[1] = [[NSColor colorWithCalibratedRed:1.0 green:1.0 blue:0.4 alpha:1.0] retain]; // light yellow
    colors[2] = [[NSColor colorWithCalibratedRed:1.0 green:0.6 blue:0.6 alpha:1.0] retain]; // light pink
    colors[3] = [[NSColor colorWithCalibratedRed:0.8 green:1.0 blue:0.4 alpha:1.0] retain]; // light green
    colors[4] = [[NSColor redColor] retain]; // hot red.


    return self;
}

- (void)loadDefaults
{
    speed = [[[NSUserDefaults standardUserDefaults] objectForKey:@"FreeSpaceSpeed"] intValue];
    if ( (speed < 1) || (speed > 5) ) speed = 3;

    stars = [[[NSUserDefaults standardUserDefaults] objectForKey:@"FreeSpaceStars"] intValue];
    if ( (stars < 1) || (stars > 5) ) stars = 3;

    color = [[[NSUserDefaults standardUserDefaults] objectForKey:@"FreeSpaceColor"] isEqualToString:@"YES"];
}

- (NSView *)inspector:sender
{
    if (!sharedInspectorPanel) {
        [NSBundle loadNibNamed:@"FreeSpaceView.nib" owner:self];
    }

    return sharedInspectorPanel;
}

//
// The Graphics Code
//

- (void)drawRect:(NSRect)rects
{
    // blank out area
    [[NSColor blackColor] set];
    NSRectFill(rects);
}

- (void)oneStep
{
    [self moveStars:YES];
    [self drawStars];
}

- (void)moveStars:(BOOL)prepDrawing
{
    int i;
    Star *s;
    NSRect bounds = [self bounds];

    drawRects = 0;

    for (i = 0; i < stars * starsMultiplier; i++) {
        s = &starList[i];

        // move the point, accelerate
        s->distance += s->d;
        s->d *= s->dd;
        s->drawpt = drawptForStarInRect(s, bounds);

        // hack for spinning stars:
        /*s->theta += 0.01;
        s->costheta = cos(s->theta);
        s->sintheta = sin(s->theta);*/

        // create the erasure list entry.
        if (prepDrawing){
            eraseRectList[drawRects].size = starsSizes[s->changeIndex];
            eraseRectList[drawRects].origin = s->erasept;
        }

        // if our star has moved off the screen, create a new star
        if ( (s->drawpt.x < 0) || (s->drawpt.y < 0) ||
             (s->drawpt.x + 7 > bounds.size.width) ||
             (s->drawpt.y + 7 > bounds.size.height) )  [self replaceStar:i];

         // create the draw list entry
        if (prepDrawing) {
            while (s->distance > s->changeList[s->changeIndex]) {
                s->changeIndex++; // this is a little dangerous.
                if (s->changeIndex > 6) {
                    s->changeIndex = 6;
                    NSLog(@"changeIndex overflow ... corrected.");
                }
            }
            drawRectList[drawRects].size = starsSizes[s->changeIndex];
            drawRectList[drawRects].origin = s->drawpt;
            drawRects++;
        }

        // set the new erase point.
        s->erasept = s->drawpt;
    }
}

- (void)drawStars
{
    int j = 0, cnt;
    NSRect *newDList,*newEList;

    // Modern systems are quite capable of pumping out 800 stars per frame, but the time delay between the first
    // black square and the first non-black squares is big enough to cause 'flicker'. The original Space module
    // broke drawing up into 100 star chunks; we do that same thing here. Once machines can eat 800 stars for
    // lunch, the following commented out block and all of our local variables can be removed.

    /*[[NSColor blackColor] set];
    NSRectFillList(eraseRectList, drawRects);
    if (!color) {
        [[NSColor whiteColor] set];
        NSRectFillList(drawRectList, drawRects);
    } else {
        int i;
        int num = drawRects / kMaxColors;

        for (i = 0; i < kMaxColors; i++) {
            [colors[i] set];
            NSRectFillList(&drawRectList[i * num], num);
        }
    }*/

    do {
        cnt = (j + kMaxRectRedraw > drawRects) ? drawRects - j : kMaxRectRedraw;
        if (cnt == 0) break; // shouldn't be.
        newEList = &eraseRectList[j];
        newDList = &drawRectList[j];

        [[NSColor blackColor] set];
        NSRectFillList(newEList, cnt);
        if (!color) {
            [[NSColor whiteColor] set];
            NSRectFillList(newDList, cnt);
        } else {
            int i;
            int num = cnt / kMaxColors;

            for (i = 0; i < kMaxColors; i++) {
                [colors[i] set];
                NSRectFillList(&newDList[i * num], num);
            }
        }
        j += cnt;
    } while (j < drawRects);
}

- (void)replaceStar:(int)index
{
    Star *s = &starList[index];
    float dist;
    BOOL inBounds;
    NSRect bounds = [self bounds];

    // set the distance; make sure the point is in-bounds.
    s->distance = radius;
    s->theta = randBetween(0,(2*M_PI));
    s->costheta = cos(s->theta);
    s->sintheta = sin(s->theta);
    do {
            s->distance = randBetween(1, s->distance);

            s->drawpt = drawptForStarInRect(s,[self bounds]);

            inBounds = YES;
            if ( (s->drawpt.x < 0) || (s->drawpt.y < 0) ||
                 (s->drawpt.x + 7 > bounds.size.width) ||
                 (s->drawpt.y + 7 > bounds.size.height) )  inBounds = NO;

    } while (!inBounds);

    // set our delta and acceleration
    s->d = (0.2);
    s->dd = randBetween(MAX(1.0, 0.92 + 0.04 * (speed -1)), 1.02 + (0.04 * (speed -1))); // differences in speed are pretty subtle

    // build our changeList; loop unrolled.
    dist = MAX(20,randBetween(0, (0.42*radius)));
    s->changeList[0] = s->distance + 5;
    s->changeList[1] = s->changeList[0] - 5 + dist + dist;
    s->changeList[2] = s->changeList[1] + dist;
    s->changeList[3] = s->changeList[2] + dist;
    s->changeList[4] = s->changeList[3] + dist;
    s->changeList[5] = s->changeList[4] + dist;	
    s->changeList[6] = 1000000.0;	// an arbitrarily large number.
    s->changeIndex = 0;
}

- (void)setFrameSize:(NSSize)newSize
{
    [super setFrameSize:newSize];
    [self seedImage];
}

- (void)seedImage
{
    int i;
    float x = [self bounds].size.width;
    float y = [self bounds].size.height;

    starsMultiplier = kMaxStars / 5;

    radius = (sqrt(x*x + y*y))/2;

    // create our stars
    for (i = 0; i < stars * starsMultiplier; i++) {
        [self replaceStar:i];
    }

    // seed our stars with better initial positions.
    for (i = 0; i < 50; i++) {
        [self moveStars:NO];
    }
}


//
// User Interface
//

- (void)colorSet:(id)sender
{
    color = [colorBtn state];
    [[NSUserDefaults standardUserDefaults] setObject:(color?@"YES":@"NO") forKey:@"FreeSpaceColor"];
}

- (void)speedSet:(id)sender
{
    speed = [speedSlider intValue];
    [speedText setStringValue:speedStrings[speed-1]];

    [[NSUserDefaults standardUserDefaults] setObject:[NSString stringWithFormat:@"%d",speed] forKey:@"FreeSpaceSpeed"];
    [self seedImage];
}

- (void)starsSet:(id)sender
{
    stars = [starsSlider intValue];
    [starsText setStringValue:starsStrings[stars-1]];

    [[NSUserDefaults standardUserDefaults] setObject:[NSString stringWithFormat:@"%d",stars] forKey:@"FreeSpaceStars"];
    [self seedImage];
}

@end

@implementation StaticFreeSpaceView

- (id)initWithFrame:(NSRect)frameRect
{
    if (![self initWithFrame:frameRect loadInspectorPanel:NO]) return nil;

    speed = 3;
    stars = 2;

    [self seedImage];
    [self moveStars:YES];

    return self;
}

- (void)drawRect:(NSRect)rects
{
    // clip to our rect.
    NSRectClip(rects);

    // do the drawing, which is now clipped to 'rects'
    [super drawRect:rects];
    [self drawStars];
}


@end
