/*
 * Messages.m
 * 7/6/98 Mike Trent
 *
 * Messages screen saver module.
 *
 * 8/10/98  MDT added setFrameSize to better deal with BackSpace.
 */

#import "MessagesView.h"

static NSString *speedStrings[] = {
    @"Slowest",
    @"Slow",
    @"Medium",
    @"Fast",
    @"Fastest",
};

enum {
    kMarquee = 0,
    kMovieCredits,
    kRicochet,
    kStationary,
};

// defined by FreeSpace
float randBetween(float a, float b);

static float randomSign(void)
{
    return ( (random() - (RAND_MAX/2)) > 0 ) ? 1.0 : -1.0;
}

@implementation MessagesView

//
// Initialization
//

- (BOOL)useBufferedWindow
{
    return NO;
}

- (id)initWithFrame:(NSRect)frameRect
{
    return [self initWithFrame:frameRect loadInspectorPanel:YES];
}

- (id)initWithFrame:(NSRect)frameRect loadInspectorPanel:(BOOL)loadPanel
{
    if (![super initWithFrame:frameRect]) return nil;

    [self allocateGState];
    srandom(time(0));

    [self loadDefaults];

    if (loadPanel) {
        (void)[self inspector:self];

        // set defaults
        [styleMenu selectItemWithTitle:styleStr];
        style = [styleMenu indexOfSelectedItem];
        [speedSlider setIntValue:speed];
        [speedText setStringValue:speedStrings[speed-1]];
        [textField setString:text];
        [textField setDelegate:self]; // can't do this in IB for some reason.

        [textWindow setFrameAutosaveName:@"MessagesTextWindow"];
   }

    return self;
}

- (void)loadDefaults
{
    styleStr = [[NSUserDefaults standardUserDefaults] objectForKey:@"MessagesStyle"];
    if ((!styleStr) || ([styleStr isEqualToString:@""])) styleStr = @"Stationary";
    [styleStr retain];

    speed = [[[NSUserDefaults standardUserDefaults] objectForKey:@"MessagesSpeed"] intValue];
    if ( (speed < 1) || (speed > 5) ) speed = 3;

    text = [[NSUserDefaults standardUserDefaults] objectForKey:@"MessagesText"];
    if ((!text) || ([text isEqualToString:@""])) text = @"Your Text Goes Here!";
    [text retain];
}

- (NSView *)inspector:sender
{
    if (!sharedInspectorPanel) {
        [NSBundle loadNibNamed:@"Messages.nib" owner:self];
    }

    return sharedInspectorPanel;
}

- (void)inspectorWillBeRemoved
{
    [textWindow close];
}

//
// The Graphics Code
//

- (void)drawRect:(NSRect)rects
{
    // blank out area
    [[NSColor blackColor] set];
    NSRectFill(rects);

    // if we have an image (which we should) and we need to redraw part of our image, redraw the image
    // I guess I could use compositeToPoint:fromRect:operation to only redraw the part that needs to be
    // redrawn, but I'm lazy.
    if (image) {
        NSRect r;

        r.origin = pos;
        r.size = [image size];

        if (!NSIsEmptyRect(NSIntersectionRect(r , rects))) {
            [image compositeToPoint:pos operation:NSCompositeCopy];
        }
    }
}

- (void)oneStep
{
    NSRect r;
    NSRect bounds = [self bounds];

    pos.x += (d.x * speed);
    pos.y += (d.y * speed);
    r.size = [image size];
    r.origin = pos;

    switch (style) {
    case kMarquee:
    case kMovieCredits:
        // If we're moving offscreen, erase our region and start again.
        if (NSIsEmptyRect(NSIntersectionRect(r , bounds))) {
            [[NSColor blackColor] set];
            NSRectFill(r);
            [self createTextImage:[self bounds]];
        }
        break;
    case kRicochet:
        // If we're moving offscreen, 'bounce'.
        if ((r.origin.x < bounds.origin.x) || (r.origin.x + r.size.width > bounds.origin.x + bounds.size.width))
            d.x *= -1;
        if ((r.origin.y < bounds.origin.y) || (r.origin.y + r.size.height > bounds.origin.y + bounds.size.height))
            d.y *= -1;
        break;
    case kStationary:// fall through
    default:
        if ( (!timer) || ([timer compare:[NSDate date]] == NSOrderedAscending) )  {
            [[NSColor blackColor] set];
            NSRectFill(r);
            [self createTextImage:[self bounds]];
        }
    }

    [image compositeToPoint:pos operation:NSCompositeCopy];
}

- (void)setFrameSize:(NSSize)newSize
{
    [super setFrameSize:newSize];
    [self createTextImage:[self bounds]];
}

- (void)willExitScreenSaverMode
{
    [image release];
    image = nil;
    [timer release];
    timer = nil;
}

- (void)createTextImage:(NSRect)bounds
{
    NSCell *cell;
    NSRect bigRect = bounds;
    NSSize imageSize;
    NSRect imageRect = {{0,0 },{0,0 } };
    NSRect textRect = {{10,10 },{0,0 } };

    // create our text cell.
    cell = [self cellForString:text color:[self randomColor] font:[self randomFont] wrap:(style != kMarquee)];

    // divine our image size, position point, and delta point. This is potentially different for each kind of animation.
    switch (style) {
        case kMarquee:
            bigRect.size.width = 99999.0; // arbitrarily large
            imageSize = [cell cellSizeForBounds:bigRect];
            textRect.size = imageSize;
            imageSize.height += 20; imageSize.width += 20;  // we're padding our image 10 pixels on a side for easy re-drawing
            imageRect.size = imageSize;
            pos = [self randomPointForSize:imageSize within:bounds];
            pos.x = bounds.size.width - 1;
            d = NSMakePoint(-1 ,0);
            break;
        case kMovieCredits:
            bigRect.size.height = 99999.0; // arbitrarily large
            imageSize = [cell cellSizeForBounds:bigRect];
            textRect.size = imageSize;
            imageSize.height += 20; imageSize.width += 20;
            imageRect.size = imageSize;
            pos = [self randomPointForSize:imageSize within:bounds];
            pos.y = 1-imageSize.height;
            d = NSMakePoint(0, 1);
            break;
        case kRicochet:
            imageSize = [cell cellSizeForBounds:bounds];
            textRect.size = imageSize;
            imageSize.height += 20; imageSize.width += 20;
            imageRect.size = imageSize;
            pos = [self randomPointForSize:imageSize within:bounds];
            d = NSMakePoint(randomSign(), randomSign());
            break;
        case kStationary:// fall through
        default:
            imageSize = [cell cellSizeForBounds:bounds];
            textRect.size = imageSize;
            textRect.origin = NSMakePoint(0, 0);
            imageRect.size = imageSize;
            pos = [self randomPointForSize:imageSize within:bounds];
            d = NSMakePoint(0, 0);
            timer = [[NSDate dateWithTimeIntervalSinceNow:(12.0 * speed)] retain];
    }

    // we'll be drawing in black today ...
    [[NSColor blackColor] set];

    // create the image.
    if (image) [image release];
    image = [[NSImage alloc] initWithSize:imageSize];
    [image lockFocus];
    NSRectFill(imageRect);
    [cell drawWithFrame:textRect inView:[NSView focusView]];
    [image unlockFocus];

    // blank out the whole view, just in case. Our blackColor is still set.
    //
    // MDT 2/24/99 Looks like this is no longer necessary. What's more, it's
    // causing problems.
    // NSRectFill(bounds);
}

//
// User Interface
//

- (void)speedSet:(id)sender
{
    speed = [speedSlider intValue];
    [speedText setStringValue:speedStrings[speed-1]];

    [[NSUserDefaults standardUserDefaults] setObject:[NSString stringWithFormat:@"%d",speed] forKey:@"MessagesSpeed"];
}

- (void)styleSet:(id)sender
{
    if (styleStr) [styleStr release];
    styleStr = [[styleMenu titleOfSelectedItem] retain];
    style = [styleMenu indexOfSelectedItem];
    [[NSUserDefaults standardUserDefaults] setObject:styleStr forKey:@"MessagesStyle"];
    [self createTextImage:[self bounds]];
}

- (void)textSet:(id)sender
{
    [textWindow makeKeyAndOrderFront:nil];
}

- (void)textDidBeginEditing:(NSNotification *)aNotification
{
    if (text) [text release];
    text = [[textField string] retain];
    [[NSUserDefaults standardUserDefaults] setObject:text forKey:@"MessagesText"];
    [self createTextImage:[self bounds]];
}

//
// Utility
//

- (NSPoint)randomPointForSize:(NSSize)ssize within:(NSRect)bounds
{
    NSPoint pt;

    pt.x = floor(randBetween(bounds.origin.x, bounds.size.width - ssize.width));
    pt.y = floor(randBetween(bounds.origin.y, bounds.size.height - ssize.height));

    return pt;
}

/*
 * Code contributed by Mike Ferris, modified by me (Mike Trent)
 * I should really move this into a Library ...
 */

- (NSCell *)cellForString:(id)str color:(NSColor *)color font:(NSFont *)font wrap:(BOOL)wrap
{
    // str can be NSString or NSAttributedString.

    static NSTextFieldCell *cell = nil;
    if (!cell) {
        cell = [[NSTextFieldCell allocWithZone:[self zone]] initTextCell:@"Hello"];
        [cell setBordered:NO];
        [cell setBezeled:NO];
        [cell setDrawsBackground:NO];
    }
    if ([str isKindOfClass:[NSAttributedString class]]) {
        [cell setAttributedStringValue:str];

        // MF:??? The color of text in rich text documents is usually black.  This is bad.  What should be done?
        // MF:??? The font of text in rich text documents is usually too small for this purpose.  What should be done?
    } else {
        [cell setWraps:wrap]; // MDT
        [cell setStringValue:str];
        [cell setFont:font];
        [cell setTextColor:color];
    }

    return cell;
}

- (NSColor *)randomColor
{
    float red, green, blue;

    red = ((float)random()) / (float)((unsigned)((unsigned)(1<<31) - 1));
    green = ((float)random()) / (float)((unsigned)((unsigned)(1<<31) - 1));
    blue = ((float)random()) / (float)((unsigned)((unsigned)(1<<31) - 1));
    return [NSColor colorWithCalibratedRed:red green:green blue:blue alpha:1.0];
}

- (NSFont *)randomFont
{
    static NSMutableArray *availableFonts = nil;
    unsigned count, index;
    NSFont *font = nil;
    NSString *name;

    if (!availableFonts) {
        availableFonts = [[[NSFontManager sharedFontManager] availableFonts] mutableCopyWithZone:[self zone]];
    }
    count = [availableFonts count];

    while (count > 0) {
        index = (random() % (long)count);
        name = [availableFonts objectAtIndex:index];
        font = [NSFont fontWithName:name size:[self randomFontSize]]; // MDT a nice random size (TODO: make better)

        // If the font is NeXTDoc or Lexi or if it has a FontSpecific encoding, skip it.
        // MF:!!! This hardly seems like a general strategy...
        if ([name isEqualToString:@"NeXTdoc"] || [name isEqualToString:@"Lexi"] || [[font encodingScheme] isEqualToString:@"FontSpecific"]) {
            [availableFonts removeObjectAtIndex:index];
            count--;
        } else {
            break;
        }
    }

    return font;
}

// MDT
- (float)randomFontSize
{
    return (54.0 - 36.0) * (random() / RAND_MAX) + 36.0;
}

@end

@implementation StaticMessagesView

- (id)initWithFrame:(NSRect)frameRect
{
    NSRect r,bounds;
    int saftey=0;

    if (![self initWithFrame:frameRect loadInspectorPanel:NO]) return nil;

    style = kStationary;
    bounds = [self bounds];
    r = NSMakeRect(bounds.origin.x - (bounds.size.width/2),
                   bounds.origin.y - (bounds.size.height/2),
                   bounds.size.width * 2,
                   bounds.size.height* 2);

    do {
        [self createTextImage:r];
        [timer release];
        r.origin = pos;
        r.size = [image size];
        if (saftey++ > 100) break;
    } while  (NSIsEmptyRect(NSIntersectionRect([self bounds] , r)));

    return self;
}

- (NSColor *)randomColor
{
    return [NSColor grayColor];
}

- (NSCell *)cellForString:(id)str color:(NSColor *)color font:(NSFont *)font wrap:(BOOL)wrap
{
    return [super cellForString:str color:color font:font wrap:NO];
}

- (void)drawRect:(NSRect)rects
{
    NSRectClip(rects);
    [super drawRect:rects];
}

@end