/*
 * RandomizerView.m
 * 9/15/98 Mike Trent
 *
 * Randomizer screen saver module.
 */

#import "RandomizerView.h"

@implementation RandomizerView

//
// Initialization
//

- (id)initWithFrame:(NSRect)frameRect
{
    if (![super initWithFrame:frameRect]) return nil;

    [self allocateGState];
    srandom(time(0));

    appController = [NSApp delegate];

    [self loadDefaults];
    [self getModuleList];

    return self;
}

- (void)getModuleList
{
    NSMutableDictionary *tmp;
    
    [modules release];
    [sortedModuleNames release];
    modules = [appController modules];
    tmp = [NSMutableDictionary dictionaryWithDictionary:modules];
    [tmp  removeObjectForKey:@"Randomizer"];
    sortedModuleNames = [[[tmp allKeys] sortedArrayUsingSelector:@selector(compare:)] retain];
}

- (void)loadDefaults
{
    skipList = [[NSMutableArray arrayWithArray:[[NSUserDefaults standardUserDefaults] objectForKey:@"RandomizerSkipList"]] retain];

    debug = [[[NSUserDefaults standardUserDefaults] objectForKey:@"RandomizerDebug"] isEqualToString:@"YES"];
}

- (NSView *)inspector:sender
{
    if (!sharedInspectorPanel) {
        NSImageCell *cell = [[NSImageCell alloc] init];

        [NSBundle loadNibNamed:@"Randomizer.nib" owner:self];

        check = [[NSImage alloc] initWithContentsOfFile:[[NSBundle bundleForClass:[self class]] pathForImageResource:@"check"]];
        if (!check) NSLog(@"can't find check.tiff");
        [[table tableColumnWithIdentifier:@"On"] setDataCell:cell];
        [table setDoubleAction:@selector(doubleClick:)];
        [table setTarget:self];
        [table setDataSource:self];
        [table setDelegate:self];
    }

    return sharedInspectorPanel;
}

//
// The Module Code
//

- (void)setRandomModule
{
    int count, index;
    NSMutableDictionary *mod;
    NSString 	*name;
    //int failsafe = 0;
    NSMutableArray *list;
    
    module = nil;

    // prune out unwanted modules. See my official apology in main comments.
    list = [NSMutableArray arrayWithArray:sortedModuleNames];
    [list removeObjectsInArray:skipList];

    // pick one.
    count = [list count];
    index = count * ((float) random() / (float) RAND_MAX);
    if (index == count) index = 0;
    name = [list objectAtIndex:index];

    // load the module.
    [appController loadModuleBundle:name];
    mod = [modules objectForKey:name];
    if (!mod) {
        NSLog(@"Can't find random module: %@", [sortedModuleNames objectAtIndex:index]);
        return;
    }

    module = [mod objectForKey:@"Object"];
    if (debug) NSLog(@"module = %@", name);

    // set subview
    [self addSubview:module];
}

- (void)setFrameSize:(NSSize)newSize
{
    [super setFrameSize:newSize];
    [module setFrameSize:newSize];
}

- (void)drawRect:(NSRect)rects
{
    if (module) [module drawRect:rects];
    else {
        [[NSColor blackColor] set];
        NSRectFill(rects);
    }
}

- (void)oneStep
{
    [module oneStep];
}

- (NSTimeInterval)animationDelayTime
{
    NSTimeInterval result = 0.04;

    if ([module respondsToSelector:@selector(animationDelayTime)])
        result = [module animationDelayTime];

    return result;
}

- (void)didLockFocus
{
    if ([module respondsToSelector:@selector(didLockFocus)])
        [module didLockFocus];
}

- (BOOL)useBufferedWindow
{
    BOOL result = NO;

    if ([module respondsToSelector:@selector(useBufferedWindow)])
        result = [module useBufferedWindow];

    return result;
}

// Not Yet Implemented
- (BOOL)ignoreMouseMovement;
{
    BOOL result = NO;

    if ([module respondsToSelector:@selector(ignoreMouseMovement)])
        result = [module ignoreMouseMovement];

    return result;
}

- (void)willEnterScreenSaverMode
{
    if ([module respondsToSelector:@selector(willEnterScreenSaverMode)])
        [module willEnterScreenSaverMode];
}

- (void)enteredScreenSaverMode
{
    if ([module respondsToSelector:@selector(enteredScreenSaverMode)])
        [module enteredScreenSaverMode];
}

- (void)willExitScreenSaverMode
{
    if ([module respondsToSelector:@selector(willExitScreenSaverMode)])
        [module willExitScreenSaverMode];
}

- (BOOL)useMultipleMonitors
{
    if ([module respondsToSelector:@selector(useMultipleMonitors)])
        return [module useMultipleMonitors];

    return YES;
}

// firstCall is the first message sent to an object when the screen saver
// cycle begins. Isn't that nice?
- (void)firstCall
{
    [self setRandomModule];

    if ([module respondsToSelector:@selector(firstCall)])
        [module firstCall];
}

//
// User Interface
//


//
// NSTableView Delegate methods
//

- (int)numberOfRowsInTableView:(NSTableView *)view
{
    return [sortedModuleNames count];
}

- (id)tableView:(NSTableView *)view
            objectValueForTableColumn:(NSTableColumn *)theColumn
            row:(int)rowIndex
{
    NSParameterAssert(rowIndex >= 0 && rowIndex <= [sortedModuleNames count]);
    if ([[theColumn identifier] isEqualToString:@"Module"]) {
        return [sortedModuleNames objectAtIndex:rowIndex];
    } else if ([[theColumn identifier] isEqualToString:@"On"]) {
        if ([skipList containsObject:[sortedModuleNames objectAtIndex:rowIndex]]) {
            return nil;
        } else {
            return check;
        }
    }
    return nil;
}

- (id)doubleClick:(id)sender
{
    NSTableColumn *col;
    int index = [sender clickedRow];
    NSString *name = [sortedModuleNames objectAtIndex:index];

    col = [[sender tableColumns] objectAtIndex:[sender clickedColumn]];

    if ([skipList containsObject:name]) {
        [skipList removeObject:name];
    } else {
        [skipList addObject:name];
    }
    [sender setNeedsDisplayInRect:[sender rectOfRow:index]];

    [[NSUserDefaults standardUserDefaults] setObject:skipList forKey:@"RandomizerSkipList"];
    
    return self;
}

@end

@implementation StaticRandomizerView

- (void)drawRect:(NSRect)rects
{
    NSRectClip(rects);
    [[NSColor blackColor] set];
    NSRectFill(rects);
}

@end
