/*
 * SlideShowView.m
 * 7/1/98 Mike Trent
 *
 * SlideShow screen saver module.
 */

#import "SlideShowView.h"

static NSPoint CenterRectInRect(NSSize a, NSRect b)
{
    NSPoint p;

    p.x = (b.size.width - a.width) / 2 + b.origin.x;
    p.y = (b.size.height - a.height)/2 + b.origin.y;
    return p;
}

@implementation SlideShowView

//
// Initialization
//

- (id)initWithFrame:(NSRect)frameRect
{
    if (![super initWithFrame:frameRect]) return nil;

    [self allocateGState];
    srandom(time(0));

    [self loadDefaults];
    [self buildDirectoryList];
    [self checkDirectoryList];
    (void)[self inspector:self];
    [randomBtn setState:rnd];
    [self randomSet:self];
    [descendBtn setState:descend];
    [self descendSet:self];
    [delaySld setIntValue:delay];
    [self delaySet:self];

    [dirBrowser setAcceptsArrowKeys:YES];
    [[dirBrowser cellPrototype] setFont:[NSFont fontWithName:@"Times-Medium" size:12.0]];

    loadGraphics = NO;
    lastFile = -1;

    return self;
}

- (void)buildDirectoryList
{
    NSEnumerator *e;
    NSString	 *dir;

    if (!dirList) dirList = [[NSMutableArray alloc] initWithCapacity:5];
    [dirList removeAllObjects];

    e = [[[directories allKeys] sortedArrayUsingSelector:@selector(compare:)] objectEnumerator];
    while (dir = [e nextObject]) {
        [dirList addObject:[dir lastPathComponent]];
    }
}

- (void)checkDirectoryList
{
    NSEnumerator *e;
    NSString	 *dirName, *dir;
    BOOL	isDir;

    // make sure each dir exists. If not, NSLog an error. We won't remove them from the defaults,
    // as the directory may exist on other systems.
    e = [dirList objectEnumerator];
    while (dirName = [e nextObject]) {
        dir = [directories objectForKey:dirName];
        if ( (![[NSFileManager defaultManager] fileExistsAtPath:dir isDirectory:&isDir]) ||
             (!isDir) ) {
            NSLog(@"Cannot find directory: %@", dir);
        }
    }
}

- (NSView *)inspector:sender
{
    if (!sharedInspectorPanel) {
        [NSBundle loadNibNamed:@"SlideShow.nib" owner:self];
    }

    return sharedInspectorPanel;
}

- (void)loadDefaults
{
    rnd = [[[NSUserDefaults standardUserDefaults] objectForKey:@"SlideShowRandom"] isEqualToString:@"YES"];
    descend = [[[NSUserDefaults standardUserDefaults] objectForKey:@"SlideShowDescend"] isEqualToString:@"YES"];
    directories = [[NSMutableDictionary dictionaryWithDictionary:
        [[NSUserDefaults standardUserDefaults] objectForKey:@"SlideShowDirs"]] retain];
    delay = [[[NSUserDefaults standardUserDefaults] objectForKey:@"SlideShowDelay"] intValue];
    if (delay < 1) delay = 20;
}

//
// Graphics & other work.
//

- (void)drawRect:(NSRect)rects
{
    [[NSColor blackColor] set];
    NSRectFill(rects);

    if (lastFile >= 0) {
        NSImage *image;
        image = [[[NSImage alloc] initWithContentsOfFile:[files objectAtIndex:lastFile]] autorelease];
        [image compositeToPoint:CenterRectInRect([image size], [self bounds]) operation:NSCompositeCopy];
   }
}

- (NSTimeInterval)animationDelayTime
{
    return 1.0;
}

- (void)oneStep
{
    if (loadGraphics) {
        [self scanDirectories];
        loadGraphics = NO;
    } else {
        int next,cnt;
        NSImage *image;
        NSPoint p;

        if ([files count] <= 0) return;
        if ( (timer) && ([timer compare:[NSDate date]] != NSOrderedAscending) ) return;

        [timer release];

        cnt = [files count];
        if (rnd) {
            next = (int)((float)cnt * ((float)random() / (float)RAND_MAX));
            if (next == cnt) next = 0;
        } else {
            next = lastFile + 1;
            if (next >= cnt) next = 0;
        }

        [[NSColor blackColor] set];
        NSRectFill([self bounds]);
        image = [[[NSImage alloc] initWithContentsOfFile:[files objectAtIndex:next]] autorelease];
        p = CenterRectInRect([image size], [self bounds]);
        [image compositeToPoint:p operation:NSCompositeCopy];
        lastFile = next;

        timer = [[NSDate dateWithTimeIntervalSinceNow:(float)delay] retain];
    }
}

- (void)setFrameSize:(NSSize)newSize
{
    [super setFrameSize:newSize];

    loadGraphics = YES;
    lastFile = -1;
}

- (void)scanDirectories
{
    NSEnumerator *e, *fe;
    NSString	 *dir, *path, *file;
    NSFileManager *fm = [NSFileManager defaultManager];

    if (!files) files = [[NSMutableArray alloc] initWithCapacity:20];
    [files removeAllObjects];

    e = [dirList objectEnumerator];
    while (dir = [e nextObject]) {
        path = [directories objectForKey:dir];
        //[self scanDirectory:[directories objectWithKey:dir]];
        if (descend) {
            fe = [fm enumeratorAtPath:path];
        } else {
            fe = [[fm directoryContentsAtPath:path] objectEnumerator];
        }

        while (file = [fe nextObject]) {
            if ([[NSImage imageFileTypes] containsObject:[file pathExtension]]) {
                [files addObject:[path stringByAppendingPathComponent:file]];
            }
        }
    }
}

//
// User Interface
//

- (void)addImage:(id)sender
{
    static NSOpenPanel *panel=nil;
    int status;

    if (!panel) panel = [[NSOpenPanel openPanel] retain];

    [panel setTitle:@"Select directory to scan"];
    [panel setCanChooseDirectories:YES];
    [panel setCanChooseFiles:NO];
    [panel setAllowsMultipleSelection:NO];

    status = [panel runModalForDirectory:NSHomeDirectory() file:nil types:nil];
    if (status == NSOKButton) {
        NSString *key, *path;
        path = [panel filename];
        key = [path lastPathComponent];
        [directories setObject:path forKey:key];
        [dirList addObject:key];
        [dirList autorelease];
        dirList = [[NSMutableArray arrayWithArray:[dirList sortedArrayUsingSelector:@selector(compare:)]] retain];
        [dirBrowser loadColumnZero];

        [[NSUserDefaults standardUserDefaults] setObject:directories forKey:@"SlideShowDirs"];
    }

    // MDT 8/10/98 reload if changed.
    loadGraphics = YES;
    lastFile = -1;
}

- (void)removeImage:(id)sender
{
    int row = [dirBrowser selectedRowInColumn:0];
    NSString *key;

    if (row < 0) {
        NSBeep();
        return;
    }

    key = [dirList objectAtIndex:row];
    [directories removeObjectForKey:key];
    [dirList removeObjectAtIndex:row];
    [dirBrowser loadColumnZero];
    [[NSUserDefaults standardUserDefaults] setObject:directories forKey:@"SlideShowDirs"];

    // MDT 8/10/98 reload if changed.
    loadGraphics = YES;
    lastFile = -1;
}

- (void)randomSet:(id)sender
{
    rnd = [randomBtn state];
    [[NSUserDefaults standardUserDefaults] setObject:(rnd?@"YES":@"NO") forKey:@"SlideShowRandom"];
}

- (void)descendSet:(id)sender
{
    descend = [descendBtn state];
    [[NSUserDefaults standardUserDefaults] setObject:(descend?@"YES":@"NO") forKey:@"SlideShowDescend"];

    // MDT 8/10/98 reload if changed.
    loadGraphics = YES;
    lastFile = -1;
}

- (void)delaySet:(id)sender
{
    delay = [delaySld intValue];
    [delayText setStringValue:[NSString stringWithFormat:@"%d sec.", delay]];

    [[NSUserDefaults standardUserDefaults] setObject:[NSString stringWithFormat:@"%d",delay] forKey:@"SlideShowDelay"];
}
//
// NSBrowser Delegate methods
//

- (int)browser:(NSBrowser *)sender numberOfRowsInColumn:(int)column
{
    return [dirList count];
}

- (void)browser:(NSBrowser *)sender willDisplayCell:(id)cell atRow:(int)row column:(int)column
{
    [cell setStringValue:[dirList objectAtIndex:row]];
    [cell setLeaf:YES];
}

- (BOOL)browser:(NSBrowser *)sender selectRow:(int)row inColumn:(int)column
{
    return YES;
}

@end


@implementation SlideShowInspectorView

- (void)drawRect:(NSRect)rects
{
    [[NSColor blackColor] set];
    NSRectFill(rects);
}

@end