/*
 * YumActions.m
 * 2/22/99 Mike Trent
 *
 * A series of classes that make up our Yum state engine. YumAction is the
 * base class, which defines some data and accessors. Subclasses define
 * specific implementations.
 *
 * The YumView will call changeState:, expired, and newActionWithState: methods.
 * All other methods are private to the classes in this file. changeState: will
 * update the view's State record, changing the delta or the distance. expired
 * tells the view to aquire a new action. newActionWithState: both autoreleases
 * the current action and returns a new action in one shot. Each action keeps
 * records a list of objects it may return in newActionWithState:; this is what
 * builds the edges in our state graph.
 */

#import "YumActions.h"

#ifdef GNUSTEP
#import <math.h>
#endif

float randBetween(float a, float b);

static double kPI_DIV_2 = 0.0;

//
// Action Interfaces
//

@interface YumReverseAction : YumAction
{}
@end

@interface YumAccelerateAction : YumAction
{}
@end

@interface YumDecelerateAction : YumAction
{}
@end

@interface YumChangeDistanceAction : YumAction
{}
@end

@interface YumChangeDistanceAndReverseAction : YumAction
{
    YumReverseAction *_reverse;
    YumChangeDistanceAction *_changeDistance;
}
@end

@interface YumChangeDistanceAndAccelerateAction : YumAction
{
    YumAccelerateAction *_accelerate;
    YumChangeDistanceAction *_changeDistance;
}
@end

@interface YumChangeDistanceAndDecelerateAction : YumAction
{
    YumDecelerateAction *_decelerate;
    YumChangeDistanceAction *_changeDistance;
}
@end

@interface YumNoAction : YumAction
{}
@end

@interface YumAction (PrivateStuff)
- (void)setEndInterval:(NSTimeInterval)interval;
- (void)_printDebugMessage:(NSString*)message;
- (NSCell *)_cellForString:(id)str color:(NSColor *)color font:(NSFont *)font wrap:(BOOL)wrap;
@end

//
// Action Base Class
//

@implementation YumAction

+ (void)initialize
{
    kPI_DIV_2 = M_PI * 0.5;
}

- (id)init
{
    self = [super init];

    if (self) {
        _startDate = [[NSDate date] retain];
    }

    return self;
}

- (id)initWithState:(State*)state
{
    return [self init];
}

- (void)dealloc
{
    [_startDate release];
    [_actions release];
    [super dealloc];
}

- (BOOL)expired
{
    return _expired;
}

- (void)setEndInterval:(NSTimeInterval)interval
{
    _endInterval = interval;
}

- (void)changeState:(State*)state
{
    // note if we have expired
    NSTimeInterval now = [[NSDate date] timeIntervalSinceDate:_startDate];
    if (now >= _endInterval) _expired = YES;

    // we don't actually change the state in the super class
}

- (id)newActionForState:(State*)state
{
    if (_actions) {
        int index;
        int count = [_actions count];

        index = count * ((float) random() / (float) RAND_MAX);
        if (index == count) index = 0;

        [self autorelease];
        return [(YumAction*)[[_actions objectAtIndex:index] alloc] initWithState:state];
    }

    return self;
}

//
// Private
//

- (void)_printDebugMessage:(NSString*)message
{
    NSCell *cell;
    NSRect bigRect = [[NSView focusView] bounds];
    NSRect textRect;

    cell = [self _cellForString:message color:[NSColor whiteColor] font:[NSFont systemFontOfSize:12.0] wrap:NO];
    textRect.size = [cell cellSizeForBounds:bigRect];
    textRect.size.width = 300;
    textRect.origin = NSMakePoint(0, bigRect.size.height-textRect.size.height);
    [[NSColor blackColor] set];
    NSRectFill(textRect);
    [cell drawWithFrame:textRect inView:[NSView focusView]];
}

- (NSCell *)_cellForString:(id)str color:(NSColor *)color font:(NSFont *)font wrap:(BOOL)wrap
{
    // str can be NSString or NSAttributedString.

    static NSTextFieldCell *cell = nil;
    if (!cell) {
        cell = [[NSTextFieldCell allocWithZone:[self zone]] initTextCell:@"Hello"];
        [cell setBordered:NO];
        [cell setBezeled:NO];
        [cell setDrawsBackground:NO];
    }
    if ([str isKindOfClass:[NSAttributedString class]]) {
        [cell setAttributedStringValue:str];
    } else {
        [cell setWraps:wrap]; 
        [cell setStringValue:str];
        [cell setFont:font];
        [cell setTextColor:color];
    }

    return cell;
}

@end

//
// Yum Action Subclasses
//

@implementation YumStartAction

// YumStartAction does one revolution and then kicks off some new
// action. I'm using the delta ivars to store the original theta
// and to set a "first time through" flag.

- (id)initWithState:(State*)state
{
    self = [super init];

    if (self) {
        _sDistance = 0;
        _sDelta = state->delta;
        _eDelta = state->theta;

        // this operation will take between 5 and 12 seconds
        _endInterval = randBetween(5.0, 12.0);

        // set our action list
        _actions = [[NSArray arrayWithObjects:
            [YumNoAction class],
            [YumAccelerateAction class],
            [YumDecelerateAction class],
            [YumChangeDistanceAction class],
            [YumChangeDistanceAndReverseAction class],
            [YumChangeDistanceAndAccelerateAction class],
            [YumChangeDistanceAndDecelerateAction class],
            [YumReverseAction class],
            nil] retain];
    }

    return self;
}

- (void)changeState:(State*)state
{
    if (_expired) return;

    // if _sDistance == 0, then we're still in the first half of the rotation.
    // when theta rolls over, we set _sDistance == 1 to mark the second half
    // of the rotation.
    //
    // We do all this because we can't say _eDelta == state->theta, because
    // state->theta probably won't return to exactly the same value it was
    // before.
    if (_sDistance == 0) {
        if ( (_sDelta < 0 && state->theta > _eDelta) ||
             (_sDelta > 0 && state->theta < _eDelta)) {
            _sDistance = 1;
        }
    } else {
        if ( (_sDelta < 0 && state->theta < _eDelta) ||
             (_sDelta > 0 && state->theta > _eDelta)) {
            _expired = YES;
        }
    }
}

@end

@implementation YumReverseAction

// YumReverseAction will cause our icon wheel to reverse directions over a
// period of time. It changes a state's 'delta' field from x to -(x) linearly.

- (id)initWithState:(State*)state
{
    self = [super init];

    if (self) {
        // reverse the polarity of the neutron flow
        _sDelta = state->delta;
        _eDelta = -(state->delta);

        // this operation will take between 5 and 12 seconds
        _endInterval = randBetween(5.0, 12.0);
        
        // set our action list
        _actions = [[NSArray arrayWithObjects:
            [YumNoAction class],
            [YumAccelerateAction class],
            [YumDecelerateAction class],
            [YumChangeDistanceAction class],
            [YumChangeDistanceAndAccelerateAction class],
            [YumChangeDistanceAndDecelerateAction class],
            nil] retain];
    }

    return self;
}

- (void)changeState:(State*)state
{
    [super changeState:state];

    if (_expired) {
        state->delta = _eDelta;
    } else {
        NSTimeInterval now = [[NSDate date] timeIntervalSinceDate:_startDate];
        state->delta = _sDelta + (_eDelta - _sDelta) * (now / _endInterval);
    }
}

@end

@implementation YumAccelerateAction

// YumReverseAction will cause our icon wheel to spin faster over a period
// of time. It changes a state's 'delta' field from x to y linearly.

- (id)initWithState:(State*)state
{
    self = [super init];

    if (self) {
        int sign;
        
        // pick a new final delta.
        _sDelta = state->delta;
        sign = (_sDelta < 0) ? -1 : 1;
        _eDelta = state->delta += sign * randBetween(0.01, 0.05);
        if (sign * _eDelta >= 0.1) _eDelta = sign * 0.1; // MDT unhardcode limit.

        // this operation will take between 3 and 7 seconds
        _endInterval = randBetween(3.0, 7.0);

        // set our action list
        _actions = [[NSArray arrayWithObjects:
            [YumNoAction class],
            [YumChangeDistanceAction class],
            [YumChangeDistanceAndReverseAction class],
            [YumReverseAction class],
            nil] retain];
}

    return self;
}

- (void)changeState:(State*)state
{
    [super changeState:state];

    if (_expired) {
        state->delta = _eDelta;
    } else {
        NSTimeInterval now = [[NSDate date] timeIntervalSinceDate:_startDate];
        state->delta = _sDelta + (_eDelta - _sDelta) * (now / _endInterval);
    }
}

@end

@implementation YumDecelerateAction

// YumDecelerateAction will cause our icon wheel to spin slower over a period
// of time. It changes a state's 'delta' field from x to y linearly. It's
// remarkably similar to the YumAccelerateAction object.

- (id)initWithState:(State*)state
{
    self = [super init];

    if (self) {
        int sign;
        
        // pick a new final delta.
        _sDelta = state->delta;
        sign = (_sDelta < 0) ? -1 : 1;
        _eDelta = state->delta -= sign * randBetween(0.01, 0.05);
        if (sign * _eDelta <= 0.01) _eDelta = sign * 0.01; // MDT unhardcode limit.

        // this operation will take between 3 and 7 seconds
        _endInterval = randBetween(3.0, 7.0);

        // set our action list
        _actions = [[NSArray arrayWithObjects:
            [YumNoAction class],
            [YumChangeDistanceAction class],
            [YumReverseAction class],
            [YumChangeDistanceAndReverseAction class],
            nil] retain];
}

    return self;
}

- (void)changeState:(State*)state
{
    [super changeState:state];

    if (_expired) {
        state->delta = _eDelta;
    } else {
        NSTimeInterval now = [[NSDate date] timeIntervalSinceDate:_startDate];
        state->delta = _sDelta + (_eDelta - _sDelta) * (now / _endInterval);
    }
}

@end

@implementation YumChangeDistanceAction

// YumChangeDistanceAction will cause our icon wheel to expand or contract.
// It changes a state's 'distance' field from x to y over a cosine wave.

- (id)initWithState:(State*)state
{
    self = [super init];

    if (self) {
        // pick a new final distance.
        _sDistance = state->distance;
        _eDistance = randBetween(50.0,
                                 ([[NSView focusView] bounds].size.height-48) / 2);

        // this operation will take between 3 and 7 seconds
        _endInterval = randBetween(3.0, 7.0);

        // set our action list
        _actions = [[NSArray arrayWithObjects:
            [YumNoAction class],
            [YumAccelerateAction class],
            [YumDecelerateAction class],
            [YumReverseAction class],
            [YumChangeDistanceAction class],
            [YumChangeDistanceAndReverseAction class],
            [YumChangeDistanceAndAccelerateAction class],
            [YumChangeDistanceAndDecelerateAction class],
            nil] retain];
    }

    return self;
}

- (void)changeState:(State*)state
{
    [super changeState:state];

    if (_expired) {
        state->distance = _eDistance;
    } else {
        NSTimeInterval now = [[NSDate date] timeIntervalSinceDate:_startDate];
        state->distance = _sDistance + (_eDistance - _sDistance) * (cos(M_PI * (1 - now / _endInterval)) / 2 + 0.5);
    }
}

@end

@implementation YumChangeDistanceAndReverseAction

// YumChangeDistanceAndReverseAction will cause our icon wheel to expand or 
// contract and reverse directions at the same time. The object is a composite
// object containing YumChangeDistanceAction and YumReverseAction objects. This
// way distance changing or reversal behavior can change without affecting this
// specific class. 

- (id)initWithState:(State*)state
{
    self = [super init];

    if (self) {
        // init our state objects
        _reverse = [[YumReverseAction alloc] initWithState:state];
        _changeDistance = [[YumChangeDistanceAction alloc] initWithState:state];

        // this operation will take between 3 and 7 seconds
        _endInterval = randBetween(3.0, 7.0);
        [_reverse setEndInterval:_endInterval];
        [_changeDistance setEndInterval:_endInterval];

        // set our action list
        _actions = [[NSArray arrayWithObjects:
            [YumNoAction class],
            [YumAccelerateAction class],
            [YumDecelerateAction class],
            [YumChangeDistanceAction class],
            [YumChangeDistanceAndAccelerateAction class],
            [YumChangeDistanceAndDecelerateAction class],
            nil] retain];
    }

    return self;
}

- (void)dealloc
{
    [_reverse release];
    [_changeDistance release];
    [super dealloc];
}

- (void)changeState:(State*)state
{
    [super changeState:state];

    [_reverse changeState:state];
    [_changeDistance changeState:state];
}

@end

@implementation YumChangeDistanceAndAccelerateAction

// YumChangeDistanceAndAccelerateAction will cause our icon wheel to expand or
// contract and accelerate spinning at the same time. The object is a composite
// object containing YumChangeDistanceAction and YumAccelerateAction objects.
// This way distance changing or accelerating behavior can change without
// affecting this specific class. 

- (id)initWithState:(State*)state
{
    self = [super init];

    if (self) {
        // init our state objects
        _accelerate = [[YumAccelerateAction alloc] initWithState:state];
        _changeDistance = [[YumChangeDistanceAction alloc] initWithState:state];

        // this operation will take between 3 and 7 seconds
        _endInterval = randBetween(3.0, 7.0);
        [_accelerate setEndInterval:_endInterval];
        [_changeDistance setEndInterval:_endInterval];

        // set our action list
        _actions = [[NSArray arrayWithObjects:
            [YumNoAction class],
            [YumReverseAction class],
            [YumChangeDistanceAction class],
            [YumChangeDistanceAndReverseAction class],
            nil] retain];
    }

    return self;
}

- (void)dealloc
{
    [_accelerate release];
    [_changeDistance release];
    [super dealloc];
}

- (void)changeState:(State*)state
{
    [super changeState:state];

    [_accelerate changeState:state];
    [_changeDistance changeState:state];
}

@end

@implementation YumChangeDistanceAndDecelerateAction

// YumChangeDistanceAndDecelerateAction will cause our icon wheel to expand or
// contract and decelerate spinning at the same time. The object is a composite
// object containing YumChangeDistanceAction and YumDecelerateAction objects.
// This way distance changing or decelerating behavior can change without 
// affecting this specific class. 

- (id)initWithState:(State*)state
{
    self = [super init];

    if (self) {
        // init our state objects
        _decelerate = [[YumDecelerateAction alloc] initWithState:state];
        _changeDistance = [[YumChangeDistanceAction alloc] initWithState:state];

        // this operation will take between 3 and 7 seconds
        _endInterval = randBetween(3.0, 7.0);
        [_decelerate setEndInterval:_endInterval];
        [_changeDistance setEndInterval:_endInterval];

        // set our action list
        _actions = [[NSArray arrayWithObjects:
            [YumNoAction class],
            [YumReverseAction class],
            [YumChangeDistanceAction class],
            [YumChangeDistanceAndReverseAction class],
            nil] retain];
    }

    return self;
}

- (void)dealloc
{
    [_decelerate release];
    [_changeDistance release];
    [super dealloc];
}

- (void)changeState:(State*)state
{
    [super changeState:state];

    [_decelerate changeState:state];
    [_changeDistance changeState:state];
}

@end

@implementation YumNoAction

// YumNoAction basically does nothing over a period of time. It also leads to
// and from all other nodes on the graph.

- (id)initWithState:(State*)state
{
    self = [super init];

    if (self) {
        // this operation will take between 3 and 5 seconds
        _endInterval = randBetween(3.0, 5.0);

        // set our action list
        _actions = [[NSArray arrayWithObjects:
            [YumAccelerateAction class],
            [YumDecelerateAction class],
            [YumChangeDistanceAction class],
            [YumChangeDistanceAndReverseAction class],
            [YumChangeDistanceAndAccelerateAction class],
            [YumChangeDistanceAndDecelerateAction class],
            [YumReverseAction class],
            nil] retain];
}

    return self;
}

@end
